@@ -7,22 +7,33 @@ to be an installation dependency for our packages yet--it is still too unstable
 (the latest version on PyPI doesn't even install).
 """
 
+# These first two imports are not used, but are needed to get around an
+# irritating Python bug that can crop up when using ./setup.py test.
+# See: http://www.eby-sarna.com/pipermail/peak/2010-May/003355.html
+try:
+    import multiprocessing
+except ImportError:
+    pass
+import logging
+
 import os
 import re
 import sys
+import traceback
 
 from collections import defaultdict
 
 import distutils.ccompiler
 
 from distutils import log
-from distutils.errors import DistutilsOptionError, DistutilsModuleError
+from distutils.errors import (DistutilsOptionError, DistutilsModuleError,
+                              DistutilsFileError)
+from setuptools.command.egg_info import manifest_maker
 from setuptools.dist import Distribution
 from setuptools.extension import Extension
-try:
-    from ConfigParser import RawConfigParser
-except ImportError:
-    from configparser import RawConfigParser
+
+from .extern.six import moves as m
+RawConfigParser = m.configparser.RawConfigParser
 
 
 # A simplified RE for this; just checks that the line ends with version
@@ -30,6 +41,72 @@ except ImportError:
 _VERSION_SPEC_RE = re.compile(r'\s*(.*?)\s*\((.*)\)\s*$')
 
 
+# Mappings from setup() keyword arguments to setup.cfg options;
+# The values are (section, option) tuples, or simply (section,) tuples if
+# the option has the same name as the setup() argument
+D1_D2_SETUP_ARGS = {
+    "name": ("metadata",),
+    "version": ("metadata",),
+    "author": ("metadata",),
+    "author_email": ("metadata",),
+    "maintainer": ("metadata",),
+    "maintainer_email": ("metadata",),
+    "url": ("metadata", "home_page"),
+    "description": ("metadata", "summary"),
+    "keywords": ("metadata",),
+    "long_description": ("metadata", "description"),
+    "download-url": ("metadata",),
+    "classifiers": ("metadata", "classifier"),
+    "platforms": ("metadata", "platform"),  # **
+    "license": ("metadata",),
+    # Use setuptools install_requires, not
+    # broken distutils requires
+    "install_requires": ("metadata", "requires_dist"),
+    "setup_requires": ("metadata", "setup_requires_dist"),
+    "provides": ("metadata", "provides_dist"),  # **
+    "obsoletes": ("metadata", "obsoletes_dist"),  # **
+    "package_dir": ("files", 'packages_root'),
+    "packages": ("files",),
+    "package_data": ("files",),
+    "data_files": ("files",),
+    "scripts": ("files",),
+    "py_modules": ("files", "modules"),   # **
+    "cmdclass": ("global", "commands"),
+    # Not supported in distutils2, but provided for
+    # backwards compatibility with setuptools
+    "use_2to3": ("backwards_compat", "use_2to3"),
+    "zip_safe": ("backwards_compat", "zip_safe"),
+    "tests_require": ("backwards_compat", "tests_require"),
+    "dependency_links": ("backwards_compat",),
+    "include_package_data": ("backwards_compat",),
+}
+
+# setup() arguments that can have multiple values in setup.cfg
+MULTI_FIELDS = ("classifiers",
+                "platforms",
+                "install_requires",
+                "provides",
+                "obsoletes",
+                "packages",
+                "package_data",
+                "data_files",
+                "scripts",
+                "py_modules",
+                "dependency_links",
+                "setup_requires",
+                "tests_require",
+                "cmdclass")
+
+# setup() arguments that contain boolean values
+BOOL_FIELDS = ("use_2to3", "zip_safe", "include_package_data")
+
+
+CSV_FIELDS = ("keywords",)
+
+
+log.set_verbosity(log.INFO)
+
+
 def resolve_name(name):
     """Resolve a name like ``module.object`` to an object and return it.
 
@@ -37,21 +114,23 @@ def resolve_name(name):
     """
 
     parts = name.split('.')
-    cursor = len(parts)
+    cursor = len(parts) - 1
     module_name = parts[:cursor]
+    attr_name = parts[-1]
 
     while cursor > 0:
         try:
-            ret = __import__('.'.join(module_name))
+            ret = __import__('.'.join(module_name), fromlist=[attr_name])
             break
         except ImportError:
             if cursor == 0:
                 raise
             cursor -= 1
             module_name = parts[:cursor]
+            attr_name = parts[cursor]
             ret = ''
 
-    for part in parts[1:]:
+    for part in parts[cursor:]:
         try:
             ret = getattr(ret, part)
         except AttributeError:
@@ -60,8 +139,6 @@ def resolve_name(name):
     return ret
 
 
-
-# TODO: This is getting pretty hefty; might want to break it up a bit
 def cfg_to_args(path='setup.cfg'):
     """ Distutils2 to distutils1 compatibility util.
 
@@ -74,54 +151,6 @@ def cfg_to_args(path='setup.cfg'):
             When the setup.cfg file is not found.
 
     """
-    # We need to declare the following constants here so that it's easier to
-    # generate the setup.py afterwards, using inspect.getsource.
-
-    # XXX ** == needs testing
-    D1_D2_SETUP_ARGS = {"name": ("metadata",),
-                        "version": ("metadata",),
-                        "author": ("metadata",),
-                        "author_email": ("metadata",),
-                        "maintainer": ("metadata",),
-                        "maintainer_email": ("metadata",),
-                        "url": ("metadata", "home_page"),
-                        "description": ("metadata", "summary"),
-                        "long_description": ("metadata", "description"),
-                        "download-url": ("metadata",),
-                        "classifiers": ("metadata", "classifier"),
-                        "platforms": ("metadata", "platform"),  # **
-                        "license": ("metadata",),
-                        # Use setuptools install_requires, not
-                        # broken distutils requires
-                        "install_requires": ("metadata", "requires_dist"),
-                        "provides": ("metadata", "provides_dist"),  # **
-                        "obsoletes": ("metadata", "obsoletes_dist"),  # **
-                        "package_dir": ("files", 'packages_root'),
-                        "packages": ("files",),
-                        "package_data": ("files",),
-                        "data_files": ("files",),
-                        "scripts": ("files",),
-                        "py_modules": ("files", "modules"),   # **
-                        "cmdclass": ("global", "commands"),
-                        # Not supported in distutils2, but provided for
-                        # backwards compatibility with setuptools
-                        "use_2to3": ("backwards_compat", "use_2to3"),
-                        "zip_safe": ("backwards_compat", "zip_safe")
-                        }
-
-    MULTI_FIELDS = ("classifiers",
-                    "platforms",
-                    "install_requires",
-                    "provides",
-                    "obsoletes",
-                    "packages",
-                    "package_data",
-                    "data_files",
-                    "scripts",
-                    "py_modules",
-                    "cmdclass")
-
-    BOOL_FIELDS = ("use_2to3", "zip_safe")
 
     # The method source code really starts here.
     parser = RawConfigParser()
@@ -134,17 +163,77 @@ def cfg_to_args(path='setup.cfg'):
         config[section] = dict(parser.items(section))
 
     # Run setup_hooks, if configured
-    # TODO: We need a better way of displaying errors that occur in setup_hook;
-    # right now they only show up as errors 'parsing' the cfg file.  A
-    # traceback and other info would be nice...
     setup_hooks = has_get_option(config, 'global', 'setup_hooks')
-    if setup_hooks:
-        setup_hooks = split_multiline(setup_hooks)
-        for hook in setup_hooks:
-            hook = resolve_name(hook)
-            hook(config)
+    package_dir = has_get_option(config, 'files', 'packages_root')
 
-    register_custom_compilers(config)
+    # Add the source package directory to sys.path in case it contains
+    # additional hooks, and to make sure it's on the path before any existing
+    # installations of the package
+    if package_dir:
+        package_dir = os.path.abspath(package_dir)
+        sys.path.insert(0, package_dir)
+
+    try:
+        if setup_hooks:
+            setup_hooks = split_multiline(setup_hooks)
+            for hook in setup_hooks:
+                hook_fn = resolve_name(hook)
+                try :
+                    hook_fn(config)
+                except SystemExit:
+                    log.error('setup hook %s terminated the installation')
+                except:
+                    e = sys.exc_info()[1]
+                    log.error('setup hook %s raised exception: %s\n' %
+                              (hook, e))
+                    log.error(traceback.format_exc())
+                    sys.exit(1)
+
+        kwargs = setup_cfg_to_setup_kwargs(config)
+
+        register_custom_compilers(config)
+
+        ext_modules = get_extension_modules(config)
+        if ext_modules:
+            kwargs['ext_modules'] = ext_modules
+
+        entry_points = get_entry_points(config)
+        if entry_points:
+            kwargs['entry_points'] = entry_points
+
+        wrap_commands(kwargs)
+
+        # Handle the [files]/extra_files option
+        extra_files = has_get_option(config, 'files', 'extra_files')
+        if extra_files:
+            extra_files = split_multiline(extra_files)
+            # Let's do a sanity check
+            for filename in extra_files:
+                if not os.path.exists(filename):
+                    raise DistutilsFileError(
+                        '%s from the extra_files option in setup.cfg does not '
+                        'exist' % filename)
+            # Unfortunately the only really sensible way to do this is to
+            # monkey-patch the manifest_maker class
+            @monkeypatch_method(manifest_maker)
+            def add_defaults(self, extra_files=extra_files, log=log):
+                log.info('[d2to1] running patched manifest_maker command '
+                          'with extra_files support')
+                add_defaults._orig(self)
+                self.filelist.extend(extra_files)
+
+    finally:
+        # Perform cleanup if any paths were added to sys.path
+        if package_dir:
+            sys.path.pop(0)
+
+    return kwargs
+
+
+def setup_cfg_to_setup_kwargs(config):
+    """Processes the setup.cfg options and converts them to arguments accepted
+    by setuptools' setup() function.
+    """
 
     kwargs = {}
 
@@ -163,7 +252,7 @@ def cfg_to_args(path='setup.cfg'):
             # There is no such option in the setup.cfg
             if arg == "long_description":
                 in_cfg_value = has_get_option(config, section,
-                                              "description-file")
+                                              "description_file")
                 if in_cfg_value:
                     in_cfg_value = split_multiline(in_cfg_value)
                     value = ''
@@ -177,6 +266,8 @@ def cfg_to_args(path='setup.cfg'):
             else:
                 continue
 
+        if arg in CSV_FIELDS:
+            in_cfg_value = split_csv(in_cfg_value)
         if arg in MULTI_FIELDS:
             in_cfg_value = split_multiline(in_cfg_value)
         elif arg in BOOL_FIELDS:
@@ -187,7 +278,7 @@ def cfg_to_args(path='setup.cfg'):
                 in_cfg_value = False
 
         if in_cfg_value:
-            if arg == 'install_requires':
+            if arg in ('install_requires', 'tests_require'):
                 # Replaces PEP345-style version specs with the sort expected by
                 # setuptools
                 in_cfg_value = [_VERSION_SPEC_RE.sub(r'\1\2', pred)
@@ -233,15 +324,6 @@ def cfg_to_args(path='setup.cfg'):
 
         kwargs[arg] = in_cfg_value
 
-    ext_modules = get_extension_modules(config)
-    if ext_modules:
-        kwargs['ext_modules'] = ext_modules
-
-    entry_points = get_entry_points(config)
-    if entry_points:
-        kwargs['entry_points'] = entry_points
-
-    wrap_commands(kwargs)
     return kwargs
 
 
@@ -407,7 +489,8 @@ def run_command_hooks(cmd_obj, hook_kind):
                 hook_obj = resolve_name(hook)
             except ImportError:
                 err = sys.exc_info()[1] # For py3k
-                raise DistutilsModuleError(err)
+                raise DistutilsModuleError('cannot find hook %s: %s' %
+                                           (hook,err))
         else:
             hook_obj = hook
 
@@ -416,7 +499,14 @@ def run_command_hooks(cmd_obj, hook_kind):
 
         log.info('running %s %s for command %s',
                  hook_kind, hook, cmd_obj.get_command_name())
-        hook_obj(cmd_obj)
+
+        try :
+            hook_obj(cmd_obj)
+        except:
+            e = sys.exc_info()[1]
+            log.error('hook %s raised exception: %s\n' % (hook, e))
+            log.error(traceback.format_exc())
+            sys.exit(1)
 
 
 def has_get_option(config, section, option):
@@ -437,6 +527,30 @@ def split_multiline(value):
     return value
 
 
+def split_csv(value):
+    """Special behaviour when we have a comma separated options"""
+
+    value = [element for element in
+             (chunk.strip() for chunk in value.split(','))
+             if element]
+    return value
+
+
+def monkeypatch_method(cls):
+    """A function decorator to monkey-patch a method of the same name on the
+    given class.
+    """
+
+    def wrapper(func):
+        orig = getattr(cls, func.__name__, None)
+        if orig and not hasattr(orig, '_orig'):  # Already patched
+            setattr(func, '_orig', orig)
+            setattr(cls, func.__name__, func)
+        return func
+
+    return wrapper
+
+
 # The following classes are used to hack Distribution.command_options a bit
 class DefaultGetDict(defaultdict):
     """Like defaultdict, but the get() method also sets and returns the default
@@ -464,6 +578,3 @@ class IgnoreDict(dict):
         if self.__ignore.match(key):
             return
         super(IgnoreDict, self).__setitem__(key, val)
-
-
-
